/*******************************************************************
 * postest - Kyocera Scan System F PC Daemon (tested on KM-1650)
 *
 * (C) Copyright 2006 by Pavel Pisa - Originator
 *
 * The postest is distributed under the Gnu General Public License. 
 * See file COPYING for details.
 *
 ********************************************************************/

#define WITH_DAEMON
#define HAS_GETOPT_LONG
#define _GNU_SOURCE
#define PROGRAM_VERSION "0.0.3"

#include <stdio.h>
#include <pthread.h>
#include <stdlib.h>
#include <errno.h>
#include <semaphore.h>
#include <string.h>
#include <signal.h>
#include <ctype.h>
#include <getopt.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <netinet/in.h>
#include <stdio.h>
#include <arpa/inet.h>

#ifdef WITH_DAEMON
#include <syslog.h>
#endif /*WITH_DAEMON*/

#include <ul_log.h>

extern UL_LOG_CUST(ulogd_postest);

//#define UL_LOGL_DEF UL_LOGL_MSG
#define UL_LOGL_DEF UL_LOGL_DEB

ul_log_domain_t ulogd_postest = {UL_LOGL_DEF, "postest"};

/********************************************************************/

#ifdef WITH_DAEMON

int syslog_log_priorities[] = {
  [UL_LOGL_FATAL] = LOG_CRIT,
  [UL_LOGL_ERR]   = LOG_ERR,
  [UL_LOGL_MSG]   = LOG_WARNING,
  [UL_LOGL_INF]   = LOG_INFO,
  [UL_LOGL_DEB]   = LOG_DEBUG,
  [UL_LOGL_TRASH] = LOG_DEBUG,
};

void
syslog_log_fnc(ul_log_domain_t *domain, int level,
        const char *format, va_list ap)
{
  char *str;

  level = syslog_log_priorities[level & UL_LOGL_MASK];
  if(vasprintf(&str, format, ap)<0)
      return;
  if((domain == NULL) || (domain->name==NULL))
      syslog(LOG_MAKEPRI(LOG_LOCAL5, level), "%s", str);
  else
      syslog(LOG_MAKEPRI(LOG_LOCAL5, level), "%s:%s", domain->name, str);
  free(str);
}

static int syslog_log_init(void)
{
  if(getenv("UL_LOG_FILENAME")==NULL) {
    openlog ("postest", 0, LOG_LOCAL3);
    ul_log_redir(syslog_log_fnc, 0);
  }
  return 0;
}

static int daemon_init(void)
{
  pid_t pid;

  if ((pid = fork()) < 0) {
      return -1;
  } else {
      if (pid != 0) {
          exit(0);	/* parent vanishes */
      }
  }
  /* child process */

  close(0);
  close(1);
  close(2);

  setsid();
  setpgid (0, 0);
  umask(022);

  return 0;
}

#endif /*WITH_DAEMON*/

/********************************************************************/

#define HEXDUMP_ROW_MAX 32

void ul_log_hexdump(ul_log_domain_t *domain, int level, char *prefix,
                    const void *buff, unsigned long start, unsigned long len, int row,
		    int flags)
{
  static const char xdigs[] = "0123456789ABCDEF";
  unsigned long addr=start;
  unsigned char *p=(unsigned char *)buff;
  unsigned char c;
  int cnt;
  int i;
  char row_hex[HEXDUMP_ROW_MAX*3];
  char row_chr[HEXDUMP_ROW_MAX*3];
  char *p_hex;
  char *p_chr;

  if(!ul_log_cond(domain, level))
    return;

  if(!prefix)
    prefix="";

  if(!row)
    row=16;

  if(row>HEXDUMP_ROW_MAX)
    row=HEXDUMP_ROW_MAX;
  
  while(len){
    cnt=(len>row)?row:len;
    len-=cnt;
    p_hex=row_hex;
    p_chr=row_chr;
    for(i=cnt;i--;){
      c=*(p++);
      *(p_hex++)=xdigs[(c>>4)&0xf];
      *(p_hex++)=xdigs[c&0xf];
      *(p_hex++)=' ';
      if((c<0x20)||(c>0x80))
        c='.';
      *(p_chr++)=c;
    }
    *(--p_hex)=0;
    *(--p_chr)=0;
    ul_log(domain, level, "%s%04lx:%s  %s\n", prefix, addr, row_hex, row_chr);
    addr+=cnt;
  }
  return;  
}

/********************************************************************/

#define POSTEST_CALLIN_PORT  5554
#define POSTEST_CALLOUT_PORT 5553

#define CON_FDTYPE_FILE   0
#define CON_FDTYPE_SOCKET 1

typedef struct {
  int fdin;
  int fdout;
  int fdtype;
  struct sockaddr_in peer_addr;
  int errcode;
  int expected_bytes;
  int rec_bytes;
  int timeout;
  void *app_options;
} con_params_t;

typedef struct {
  char *base_dir;
  unsigned int callin_port_number;
  unsigned int callout_port_number;
} postest_app_options_t;


pthread_mutex_t file_counter_mutex=PTHREAD_MUTEX_INITIALIZER;
int file_counter;

pthread_mutex_t thread_counter_mutex=PTHREAD_MUTEX_INITIALIZER;
int thread_counter;
int thread_limit=10;

typedef struct data_item data_item_t;

struct data_item{
  char *actual_word;
  char *next_word;
  data_item_t *next;
};

pthread_rwlock_t data_items_list_mutex=PTHREAD_RWLOCK_INITIALIZER;
data_item_t *data_items_list;


pthread_mutex_t master_addr_mutex=PTHREAD_MUTEX_INITIALIZER;
struct sockaddr_in master_addr;

int data_item_add(char *actual, size_t actual_size, char *next, size_t next_size)
{
  data_item_t *item;
  item=malloc(sizeof(data_item_t));
  if(!item)
    return -1;
  item->next=NULL;
  item->actual_word=malloc(actual_size+1);
  item->next_word=malloc(next_size+1);
  if(!item->actual_word || !item->next_word){
    if(item->actual_word)
      free(item->actual_word);
    if(item->next_word)
      free(item->next_word);
    free(item);
    return -1;
  }
  memcpy(item->actual_word, actual, actual_size);
  item->actual_word[actual_size]=0;
  memcpy(item->next_word, next, next_size);
  item->next_word[next_size]=0;
  pthread_rwlock_rdlock(&data_items_list_mutex);
  item->next=data_items_list;
  data_items_list=item;
  pthread_rwlock_unlock(&data_items_list_mutex);
  return 0;
}

void data_items_list_prune(void)
{
  data_item_t *item, *p;
  pthread_rwlock_rdlock(&data_items_list_mutex);
  item=data_items_list;
  data_items_list=NULL;
  while(item){
    p=item->next;
    if(item->actual_word)
      free(item->actual_word);
    if(item->next_word)
      free(item->next_word);
    free(item);
    item=p;
  }
  pthread_rwlock_unlock(&data_items_list_mutex);
}

int con_write(con_params_t *con, const void *buff, size_t size)
{
  ssize_t retval;
  size_t left=size;

  while(left>0){
    retval=send(con->fdout,buff,left,0);
    if(retval<0){
      if(errno==EINTR)
        continue;
      ul_logmsg("con_write failed %s\n",strerror(errno));
      return -1;
    }
    left-=retval;
    buff=(char*)buff+retval;
  }
  return size-left;
}

int con_read(con_params_t *con, void *buff, size_t size, int delimchr)
{
  ssize_t retval;
  struct timeval tv;
  fd_set rfds;
  int retfd;
  size_t left=size;

  FD_ZERO(&rfds);

  while(left>0){
    tv.tv_sec = con->timeout;
    tv.tv_usec = 0;

    FD_SET(con->fdin, &rfds);
    retfd=select(con->fdin+1, &rfds, NULL, NULL, con->timeout?&tv:NULL);
    if(retfd<0){
      if(errno==EINTR)
        continue;
      ul_logmsg("con_read wait failed %s\n",strerror(errno));
      return -1;
    }

    if(retfd==0){
      ul_logmsg("con_read wait timeout\n");
      return -1;
    }

    retval=recv(con->fdin,buff,left,0);
    ul_logdeb("recv ret %d\n",retval);
    if(retval<0){
      if(errno==EINTR)
        continue;
      ul_logmsg("con_read failed %s\n",strerror(errno));
      return -1;
    }
    if(retval==0){
      ul_logmsg("con_read remote peer disconnected\n");
      return size-left;
    }
    left-=retval;
    /*ul_log_hexdump(&ulogd_postest, UL_LOGL_DEB, "recv", buff, 0, retval, 16, 0);*/
    if(delimchr>=0)
      if(memchr(buff,delimchr,retval))
        break;
    buff=(char*)buff+retval;
  }
  return size-left;
}

#define TEXT_BUFF_LEN 100

int postest_transfer(con_params_t *con)
{
  int retval=-1;
  int res;
  postest_app_options_t *options;
  long a;
  long b;
  char text_buff[TEXT_BUFF_LEN+1];
  data_item_t *item;
  
  options=(postest_app_options_t *)con->app_options;

  ul_logdeb("thread started\n");

  a=rand()&0xff;
  b=rand()&0xff;

  snprintf(text_buff,TEXT_BUFF_LEN,"%ld\n",a);
  res=con_write(con,text_buff,strlen(text_buff));
  if(res<0){
    goto finalize;
  }
  ul_logdeb("sent a\n");
  snprintf(text_buff,TEXT_BUFF_LEN,"%ld\n",b);
  res=con_write(con,text_buff,strlen(text_buff));
  if(res<0){
    goto finalize;
  }
  ul_logdeb("sent question\n");

  res=con_read(con,text_buff,TEXT_BUFF_LEN,'\n');
  if(res<0){
    goto finalize;
  }
  text_buff[res]=0;

  if(!strncmp(text_buff,"master",6)){
    char *p=text_buff+6;
    char *r,*s,*t;
    pthread_mutex_lock(&master_addr_mutex);
    if(!master_addr.sin_addr.s_addr){
      master_addr=con->peer_addr;
      res=1;
      ul_logmsg("new master set to %s\n", inet_ntoa(con->peer_addr.sin_addr));
    }else{
      res=master_addr.sin_addr.s_addr==con->peer_addr.sin_addr.s_addr?0:-1;
    }
    pthread_mutex_unlock(&master_addr_mutex);
    if(res<0){
      ul_logmsg("this is not master %s\n", inet_ntoa(con->peer_addr.sin_addr));
      goto finalize;
    }
    if(!strncmp(p,"reset",5)){
      ul_logdeb("data reset\n");
      data_items_list_prune();
    }else{
      while(*p==' ') p++;
      r=p;
      while(isalnum(*r)||(*r=='.')) r++;
      s=r;
      while(*s==' ') s++;
      t=s;
      while(isalnum(*t)||(*t=='.')) t++;
      if((s==t)||(p==r)){
        ul_logdeb("empty pair ignored\n");
        goto finalize;
      }
      data_item_add(p, r-p, s, t-s);
      ul_logdeb("new pair set\n");
    }
    retval=0;
    goto finalize;
  }

  if(atol(text_buff)!=a+b){
    ul_logdeb("bad question reply\n");
    goto finalize;
  }

  ul_logdeb("question reply OK\n");

  pthread_rwlock_rdlock(&data_items_list_mutex);

  for(item=data_items_list;item;item=item->next){
    const char delim[]="\n";
    if(con_write(con, item->actual_word, strlen(item->actual_word))<0){
      ul_logdeb("item write failed\n");
      break;
    }
    if(con_write(con, delim, strlen(delim))<0){
      ul_logdeb("delim write failed\n");
      break;
    }
  }

  do{
    int callout_fd;
    struct sockaddr_in callout_addr=con->peer_addr;
    if(callout_addr.sin_family != AF_INET)
      break;
    callout_addr.sin_port = htons(options->callout_port_number);
    callout_fd=socket(PF_INET, SOCK_STREAM, 0);
    if(callout_fd<0){
      ul_logerr("callout socket creation failed with %s\n", strerror(errno));
      break;
    }
    res=connect(callout_fd, (struct sockaddr *)&callout_addr, sizeof(struct sockaddr));
    if(res!=0){
      ul_logmsg("callout back to client failed with %s\n", strerror(errno));
      break;
    }

    for(item=data_items_list;item;item=item->next){
      const char delim[]="\n";
      if(send(callout_fd, item->next_word, strlen(item->next_word), 0)<0){
	ul_logdeb("item write failed through callout\n");
	break;
      }
      if(send(callout_fd, delim, strlen(delim), 0)<0){
	ul_logdeb("delim write failed through callout\n");
	break;
      }
    }
    close(callout_fd);
  }while(0);

  pthread_rwlock_unlock(&data_items_list_mutex);

  retval=0;

finalize:

  close(con->fdin);
  if(con->fdout!=con->fdin)
    close(con->fdout);

  ul_logmsg("postest transfer finished with %d result\n", retval);

  pthread_mutex_lock(&file_counter_mutex);
  thread_counter--;
  pthread_mutex_unlock(&file_counter_mutex);

  return retval;
}

/********************************************************************/


void *thread_scanclient(void *arg)
{
  con_params_t *con=(con_params_t *)arg;

  postest_transfer(con);

  return NULL;
}


int server_postest(int port_number, postest_app_options_t *options)
{
  int sockfd, new_fd;
  struct sockaddr_in my_addr;
  struct sockaddr_in their_addr;
  unsigned int sin_size;
  int yes = 1;
  int retval;
  pthread_t thread;
  con_params_t *con;

  sockfd = socket(PF_INET, SOCK_STREAM, 0); // do some error checking!
  if(sockfd==-1){
    ul_logerr("socket creation failed %s\n",strerror(errno));
    return -1;
  }

  if(setsockopt(sockfd, SOL_SOCKET, SO_REUSEADDR, &yes, sizeof(yes)) == -1) {
    ul_logerr("set of SO_REUSEADDR failed %s\n",strerror(errno));
    return -1;
  }
        
  memset(&my_addr, 0, sizeof(my_addr));
  my_addr.sin_family = AF_INET;
  my_addr.sin_port = htons(port_number);
  my_addr.sin_addr.s_addr = INADDR_ANY;

  retval = bind(sockfd, (struct sockaddr *)&my_addr, sizeof(struct sockaddr));
  if (retval != 0) {
    ul_logerr("bind for port %d failed %s\n",port_number, strerror(errno));
    return -1;
  }

  retval = listen(sockfd, 5);
  if (retval != 0) {
    ul_logerr("listen for port %d failed %s\n",port_number, strerror(errno));
    return -1;
  }

  ul_logmsg("server is listening for connections at port %d\n",port_number);

  while (1) {
    sin_size = sizeof(struct sockaddr_in);
    new_fd = accept(sockfd, (struct sockaddr *)&their_addr, &sin_size);

    if (new_fd < 0) {
      ul_logerr("accept for port %d failed %s\n",port_number, strerror(errno));
      return -1;
    }

    con=malloc(sizeof(con_params_t));
    if(!con){
      ul_logerr("cannot allocate connection structure\n");
      close(new_fd);
      continue;
    }

    memset(con,0,sizeof(con_params_t));
    con->fdin=new_fd;
    con->fdout=new_fd;
    con->fdtype=CON_FDTYPE_SOCKET;
    con->errcode=0;
    memcpy(&con->peer_addr, &their_addr, sizeof(struct sockaddr));
    con->app_options=options;

    ul_loginf("client connected from %s:%d\n", inet_ntoa(their_addr.sin_addr), ntohs(their_addr.sin_port));

    pthread_mutex_lock(&thread_counter_mutex);
    if(thread_counter>=thread_limit){
      pthread_mutex_unlock(&thread_counter_mutex);
      ul_logerr("concurrent clients count limit reached\n");
      close(new_fd);
      free(con);
      continue;
    }
    thread_counter++;
    pthread_mutex_unlock(&thread_counter_mutex);

    retval=pthread_create(&thread, NULL, thread_scanclient, con);
    if(retval) {
      ul_logerr("cannot create new thread %s\n", strerror(errno));
      close(new_fd);
      free(con);
      pthread_mutex_lock(&thread_counter_mutex);
      thread_counter--;
      pthread_mutex_unlock(&thread_counter_mutex);
      continue;
    }
    ul_logdeb("detaching thread\n");

    pthread_detach(thread);
  }
}

/********************************************************************/

int daemon_fl=0;
int syslog_fl=0;
int inetd_fl=0;
int set_text_fl=0;
struct sockaddr_in set_text_addr;
int set_text_range;

postest_app_options_t postest_app_options={
  .callin_port_number=POSTEST_CALLIN_PORT,
  .callout_port_number=POSTEST_CALLOUT_PORT
};

int postest_set_text(int argc, char **argv, postest_app_options_t *options)
{
  unsigned long a;
  int scan_cnt;
  struct sockaddr_in scan_addr;
  int run0=1;
  int res;
  if(!set_text_range){
    a=ntohl(set_text_addr.sin_addr.s_addr);
    set_text_range=(a|0xff)-a;
  }
  set_text_addr.sin_port=htons(options->callin_port_number);
  set_text_addr.sin_family = AF_INET;

  memcpy(&scan_addr,&set_text_addr,sizeof(scan_addr));
  scan_cnt=0;

  while(argc>0){
    int fd;

    fd=socket(PF_INET, SOCK_STREAM, 0);
    if(fd<0){
      ul_logerr("set text socket creation failed with %s\n", strerror(errno));
      break;
    }
    res=connect(fd, (struct sockaddr *)&scan_addr, sizeof(struct sockaddr));
    if(res!=0){
      ul_logdeb("set text back to client %s failed with %s\n", inet_ntoa(scan_addr.sin_addr), strerror(errno));
    }else{

      if(run0){
	const char masterreset[]="masterreset\n";
	send(fd, masterreset, strlen(masterreset), 0);
	ul_logdeb("reset succeed for %s\n", inet_ntoa(scan_addr.sin_addr));
      }else{
	const char master_[]="master ";
	if(send(fd, master_, strlen(master_), 0)>=0){
          send(fd, argv[0], strlen(argv[0]), 0);
          send(fd, " ", 1, 0);
	  if(argc>=2){
            send(fd, argv[1], strlen(argv[1]), 0);
	  }else{
            send(fd, ".", 1, 0);
	  }
          send(fd, "\n", strlen("\n"), 0);
	  argv++;
	  argc--;
          ul_logdeb("set text succeed for %s\n", inet_ntoa(scan_addr.sin_addr));
	}
      }
      sleep(1);
    }
    close(fd);
    if(++scan_cnt<set_text_range){
      a=ntohl(scan_addr.sin_addr.s_addr);
      a+=1;
      scan_addr.sin_addr.s_addr=htonl(a);
    }else{
      memcpy(&scan_addr,&set_text_addr,sizeof(scan_addr));
      scan_cnt=0;
      run0=0;
    }
  }
  return 0;
}

static void
usage(void)
{
  printf("usage: postest <parameters>\n");
 #ifdef WITH_DAEMON
  printf("  -D, --daemon             daemonize\n");
  printf("  -S, --syslog             logging to syslog\n");
  printf("  -I, --inetd              command is run from inetd\n");
 #endif /*WITH_DAEMON*/
  printf("  -p, --in-port <num>      port for incomming conn. [%d]\n",postest_app_options.callin_port_number);
  printf("  -P, --out-port <num>     port for outgoing conn. [%d]\n",postest_app_options.callout_port_number);
  printf("  -A, --out-addr <addr>    addres to set text\n");
  printf("  -R, --out-range <num>    range of address to scan\n");
  printf("  -d, --base-dir <string>  select directory where to store scans [current]\n");
  printf("  -g, --log-level <num>    default log level [%d]\n",ulogd_postest.level);
  printf("  -V, --version            print program version\n");
  printf("  -h, --help               this usage screen\n");
}

int main(int argc, char *argv[])
{

  int res;

  static const struct option long_opts[] = {
   #ifdef WITH_DAEMON
    { "daemon",   0, 0, 'D' },
    { "syslog",   0, 0, 'S' },
   #endif /*WITH_DAEMON*/
    { "int-port", 1, 0, 'p' },
    { "out-port", 1, 0, 'P' },
    { "out-addr", 1, 0, 'A' },
    { "out-range",1, 0, 'R' },
    { "base-dir", 1, 0, 'd'},
    { "log-level",1, 0, 'g'},
    { "version",  0, 0, 'V' },
    { "help",     0, 0, 'h' },
    { 0, 0, 0, 0}
  };
  static const char short_opts[]={"DSIp:P:A:R:d:g:Vh"};
  int opt;
  long l;

 #ifndef HAS_GETOPT_LONG
  while ((opt = getopt(argc, argv, short_opts)) != EOF)
 #else
  while ((opt = getopt_long(argc, argv, short_opts,
			    &long_opts[0], NULL)) != EOF)
 #endif
  switch (opt) {
    case 'D':
      daemon_fl=1;
      break;
    case 'S':
      syslog_fl=1;
      break;
    case 'p':
      postest_app_options.callin_port_number=strtoul(optarg,NULL,0);
      break;
    case 'd':
      postest_app_options.base_dir=optarg;
      break;
    case 'A':
      set_text_fl=1;
      inet_aton(optarg,&set_text_addr.sin_addr);
      break;
    case 'R':
      set_text_range=strtoul(optarg,NULL,0);
      break;
    case 'g':
      l=strtoul(optarg,NULL,0);
      if((l<0)||(l>UL_LOGL_MAX)){
        fprintf(stderr,"required log level is out of range");
	exit(3);
      }
      ulogd_postest.level=l;
      break;
    case 'V':
      fputs("postest version " PROGRAM_VERSION "\n", stdout);
      exit(0);
    case 'h':
    default:
      usage();
      exit(opt == 'h' ? 0 : 1);
  }

  {
    sigset_t my_set;
    sigemptyset(&my_set);
    sigaddset(&my_set,SIGPIPE);
    pthread_sigmask(SIG_BLOCK,&my_set,NULL);
  }

  if(set_text_fl){
    postest_set_text(argc-optind, argv+optind, &postest_app_options);
    return 0;
  }else{
    for (;optind+1 < argc; optind+=2){
      data_item_add(argv[optind], strlen(argv[optind]), argv[optind+1], strlen(argv[optind+1]));
    }
  }

#ifdef WITH_DAEMON
  if(daemon_fl) {
    syslog_log_init();
    if(daemon_init()<0){
      ul_logerr("daemonize failed\n");
      exit(2);
    }
  }else{
    if(syslog_fl)
      syslog_log_init();
  }
#endif /*WITH_DAEMON*/

  if(server_postest(postest_app_options.callin_port_number, &postest_app_options)<0){
    ul_logerr("cannot start server\n");
    exit(3);
  }

  return 0;
}
